import numpy as np
from astropy.utils.decorators import format_doc
from astropy.coordinates.representation import (
    CartesianRepresentation,
    CartesianDifferential,
    UnitSphericalRepresentation,
)
from astropy.utils import check_broadcast
from astropy.coordinates.baseframe import (
    BaseCoordinateFrame,
    base_doc,
    frame_transform_graph,
)
from astropy.coordinates.attributes import TimeAttribute
import astropy.units as un
from astropy.time import Time

from astropy.coordinates.transformations import FunctionTransformWithFiniteDifference
from astropy.coordinates.builtin_frames.icrs import ICRS

import spiceypy as spice

__all__ = ["MCMF"]

_J2000 = Time("J2000", scale="tt")


@format_doc(base_doc, components="", footer="")
class MCMF(BaseCoordinateFrame):
    """
    A coordinate or frame in the lunar "Mean Earth/ Mean Rotation frame". This is a
    "Moon-Centered/Moon-Fixed" frame, defined by an X axis through the mean position
    of the Earth-Moon direction and a Z axis through the mean rotational axis.

    In JPL ephemeris data, this is called MOON_ME.
    """

    default_representation = CartesianRepresentation
    default_differential = CartesianDifferential

    obstime = TimeAttribute(default=_J2000)

    @property
    def moon_location(self):
        """
        The data in this frame as an `~astropy.coordinates.MoonLocation` class.
        """
        from .moon import MoonLocation

        cart = self.represent_as(CartesianRepresentation)
        return MoonLocation(x=cart.x, y=cart.y, z=cart.z)


# -----------------
# Helper functions
# -----------------


def make_transform(coo, toframe):

    ap_to_spice = {"icrs": ("J2000", 0), "mcmf": ("MOON_ME", 301)}

    # Get target frame and source frame names
    from_name, from_id = ap_to_spice[coo.name]
    to_name, to_id = ap_to_spice[toframe.name]

    to_mcmf = isinstance(toframe, MCMF)
    obstime = toframe.obstime if to_mcmf else coo.obstime

    # Make arrays
    ets = np.atleast_1d((obstime - _J2000).sec)
    shape_out = check_broadcast(coo.shape, ets.shape)

    coo_cart = coo.cartesian

    mats = np.stack([spice.pxform("J2000", "MOON_ME", et) for et in ets], axis=0)
    if not to_mcmf:
        mats = np.linalg.inv(mats)

    # Check for unitspherical
    is_unitspherical = (
        isinstance(coo.data, UnitSphericalRepresentation)
        or coo.cartesian.x.unit == un.one
    )

    # If not unitspherical, shift by origin vector before rotating.
    if not is_unitspherical:
        # Make origin vector(s) in coo's frame.
        orig_posvel = (
            np.asarray([spice.spkgeo(to_id, et, from_name, from_id)[0] for et in ets])
            * un.km
        )
        coo_cart -= CartesianRepresentation((orig_posvel.T)[:3])

    newrepr = coo_cart.transform(mats).reshape(shape_out)

    if is_unitspherical:
        newrepr = newrepr.represent_as(UnitSphericalRepresentation)

    return toframe.realize_frame(newrepr)


# -----------------
# Transforms
# -----------------


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, ICRS, MCMF)
def icrs_to_mcmf(icrs_coo, mcmf_frame):
    return make_transform(icrs_coo, mcmf_frame)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, MCMF, ICRS)
def mcmf_to_icrs(mcmf_coo, icrs_frame):
    return make_transform(mcmf_coo, icrs_frame)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, MCMF, MCMF)
def mcmf_to_mcmf(mcmf_coo, mcmf_frame):
    if np.all(mcmf_coo.obstime == mcmf_frame.obstime):
        return mcmf_frame.realize_frame(mcmf_coo.data)
    else:
        # Go through ICRS to ensure new time is accounted for.
        return mcmf_coo.transform_to(ICRS()).transform_to(mcmf_frame)



from astropy.utils.data import (
    download_files_in_parallel,
    export_download_cache,
    import_download_cache,
)
import os
import sys

home = os.path.expanduser("~")
path = os.path.join(home, "ap_cache")
cache_file = os.path.join(path, "astropy_cache.zip")
os.makedirs(path, exist_ok=True)

if sys.argv[1] == "save":
    if not os.path.exists(cache_file):
        knames = [
            "lsk/naif0012.tls",
            "spk/planets/de430.bsp",
            "pck/earth_latest_high_prec.bpc",
        ]
        _naif_kernel_url = "https://naif.jpl.nasa.gov/pub/naif/generic_kernels/"
        kurls = [_naif_kernel_url + kname for kname in knames]
        paths = download_files_in_parallel(kurls, cache=True, show_progress=False)
        export_download_cache(cache_file, urls=kurls, overwrite=True)

if sys.argv[1] == "load":
    import_download_cache(cache_file)



import numpy as np
from copy import deepcopy
from lunarsky.time import Time, TimeDelta
import astropy.coordinates as ac
from astropy import units as un
from astropy.utils import IncompatibleShapeError, exceptions
from astropy.tests.helper import assert_quantity_allclose
import lunarsky
from lunarsky.moon import SELENOIDS
import pytest

try:
    from astropy.coordinates.angles.utils import angular_separation
except ImportError:
    from astropy.coordinates.angle_utilities import angular_separation

# Lunar station positions
Nangs = 7
latitudes = np.linspace(-90, 90, Nangs)
longitudes = np.linspace(0, 360, Nangs)
latlons_grid = [(lat, lon) for lon in longitudes for lat in latitudes]

# Avoiding poles:
latlons_grid_nopole = [
    (lat, lon)
    for lon in np.linspace(0.1, 360, Nangs, endpoint=False)
    for lat in np.linspace(-70.0, 70.0, Nangs, endpoint=False)
]

# Times
t0 = Time("2020-10-28T15:30:00")
_J2000 = Time("J2000")

jd_10yr = Time(t0.jd + np.linspace(0, 10 * 365.25, 5), format="jd")
jd_4mo = t0 + TimeDelta(np.linspace(0, 4 * 28 * 24 * 3600.0, 100), format="sec")


@pytest.mark.parametrize("time", jd_10yr)
@pytest.mark.parametrize("lat,lon", latlons_grid)
@pytest.mark.filterwarnings("ignore::erfa.ErfaWarning")
def test_icrs_to_topo_long_time(time, lat, lon, grcat):
    # Check that the following transformation paths are equivalent:
    #   ICRS -> MCMF -> TOPO
    #   ICRS -> TOPO
    loc = lunarsky.MoonLocation.from_selenodetic(lon, lat)
    topo0 = grcat.transform_to(lunarsky.LunarTopo(location=loc, obstime=time))
    mcmf = grcat.transform_to(lunarsky.MCMF(obstime=time))
    topo1 = mcmf.transform_to(lunarsky.LunarTopo(location=loc, obstime=time))
    assert np.all(topo0.separation(topo1) < ac.Angle("10arcsec"))


@pytest.mark.parametrize(
    "obj",
    [
        ac.get_sun(t0).transform_to(ac.ICRS),
        ac.SkyCoord(ra="30d", dec="-70d", frame="icrs"),
    ],
)
@pytest.mark.parametrize("path", [["lunartopo"], ["mcmf"], ["lunartopo", "mcmf"]])
@pytest.mark.parametrize(
    "time, lat, lon",
    [
        (t0, 11.2, 1.4),
        (t0, [10.3, 11.2], [0.0, 1.4]),
        (jd_4mo[:2], 10.3, 0.0),
        (jd_4mo[:2], [10.3, 11.2], [0.0, 1.4]),
    ],
)
@pytest.mark.parametrize("ell", SELENOIDS)
def test_transform_loops(obj, path, time, lat, lon, ell):
    # Tests from ICRS -> path -> ICRS
    obj = lunarsky.SkyCoord(obj)  # Ensure we're working with lunarsky-compatible SkyCoord
    loc = lunarsky.MoonLocation.from_selenodetic(lat, lon, ellipsoid=ell)
    fdict = {
        "lunartopo": lunarsky.LunarTopo(location=loc, obstime=time),
        "mcmf": lunarsky.MCMF(obstime=time),
    }
    orig_pos = obj.cartesian.xyz.copy()

    for fr in path:
        obj = obj.transform_to(fdict[fr])

    # Lastly, back to ICRS
    obj = obj.transform_to(ac.ICRS())
    if obj.ndim == 1:
        obj = obj[0]
    tol = 1e-4 if (obj.spherical.distance.unit == un.one) else 1 * un.m
    assert_quantity_allclose(obj.cartesian.xyz, orig_pos, atol=tol)


@pytest.mark.parametrize("ell", SELENOIDS)
def test_topo_to_topo(ell):
    # Check that zenith source transforms properly
    loc0 = lunarsky.MoonLocation.from_selenodetic(lat=0, lon=90, ellipsoid=ell)
    loc1 = lunarsky.MoonLocation.from_selenodetic(lat=0, lon=0)

    src = lunarsky.SkyCoord(alt="90d", az="0d", frame="lunartopo", location=loc0)
    new = src.transform_to(lunarsky.LunarTopo(location=loc1))
    assert new.az.deg == 90


@pytest.mark.filterwarnings("ignore::erfa.ErfaWarning")
def test_mcmf_to_mcmf():
    # Transform MCMF positions to MCMF frame half a lunar sidereal day later.
    # Assert that the new positions are roughly close to 180 deg from the original.
    src = lunarsky.SkyCoord(ra="0d", dec="0d", frame="icrs")
    src = src.transform_to(lunarsky.MCMF(obstime=jd_10yr))
    orig_pos = deepcopy(src)
    src = src.transform_to(
        lunarsky.MCMF(obstime=jd_10yr + TimeDelta(27.322 / 2, format="jd"))
    )
    sph0 = src.spherical
    sph1 = orig_pos.spherical
    res = angular_separation(sph0.lon, sph0.lat, sph1.lon, sph1.lat).to("deg")
    assert_quantity_allclose(res, 177 * un.deg, atol=5 * un.deg)
    # Tolerance to allow for lunar precession / nutation.


@pytest.mark.parametrize("ell", SELENOIDS.keys())
def test_earth_from_moon(ell):
    # Look at the position of the Earth in lunar reference frames.

    # The lunar apo/perigee vary over time. These are
    # chosen from a table of minimum/maxmium perigees over a century.
    # http://astropixels.com/ephemeris/moon/moonperap2001.html
    lunar_perigee = 356425.0  # km, Dec 6 2052
    lunar_apogee = 406709.0  # km, Dec 12 2061

    mcmf = lunarsky.spice_utils.earth_pos_mcmf(jd_4mo)
    dists = mcmf.spherical.distance.to_value("km")
    assert all((lunar_perigee < dists) & (dists < lunar_apogee))

    # # Compare position of Earth in MCMF frame from astropy and from spice
    epos_icrs_ap = ac.SkyCoord(ac.get_body_barycentric("earth", jd_4mo), frame="icrs")
    epos_icrs_sp = ac.SkyCoord(mcmf).transform_to(ac.ICRS())
    mcmf_ap = epos_icrs_ap.transform_to(lunarsky.MCMF(obstime=jd_4mo)).frame
    assert_quantity_allclose(mcmf_ap.cartesian.xyz, mcmf.cartesian.xyz, atol="6km")
    #   TODO Lower this tolerance ^

    # Whatever difference exists between the two positions should be the same in both ICRS
    # and MCMF frames:
    assert_quantity_allclose(
        np.linalg.norm(epos_icrs_ap.cartesian.xyz - epos_icrs_sp.cartesian.xyz, axis=0),
        np.linalg.norm(mcmf_ap.cartesian.xyz - mcmf.cartesian.xyz, axis=0),
    )

    # Now test LunarTopo frame positions
    lat, lon = 0, 0  # deg
    loc = lunarsky.MoonLocation.from_selenodetic(lat, lon, ellipsoid=ell)
    topo = mcmf.transform_to(lunarsky.LunarTopo(location=loc, obstime=jd_4mo))

    # The Earth should stay within 10 deg of zenith of lat=lon=0 position
    assert np.all(topo.zen.deg < 10)

    # Check that the periodicity of the Earth's motion around the zenith
    # is consistent with the Moon's orbit.
    moonfreq = 1 / (28.0 * 24.0 * 3600.0)  # Hz, frequency of the moon's orbit
    _az = np.fft.fft(topo.az.deg)
    ks = np.fft.fftfreq(jd_4mo.size, d=np.diff((jd_4mo - _J2000).sec)[0])
    sel = ks > 0

    closest = np.argmin(np.abs(moonfreq - ks[sel]))
    peakloc = np.argmax(np.abs(_az[sel]))
    assert peakloc == closest


def test_multi_times():
    # Check vectorization over time axis for LunarTopo transformations
    lat, lon = latlons_grid[6]
    loc = lunarsky.MoonLocation.from_selenodetic(lon, lat, height=10)
    star = lunarsky.SkyCoord(ra=[35], dec=[17], unit="deg", frame="icrs")
    topo0 = star.transform_to(lunarsky.LunarTopo(location=loc, obstime=jd_4mo))
    icrs0 = topo0.transform_to(ac.ICRS())

    assert np.all(star.separation(icrs0) < ac.Angle("5arcsec"))

    mcmf0 = topo0.transform_to(lunarsky.MCMF(obstime=jd_4mo))
    mcmf1 = star.transform_to(lunarsky.MCMF(obstime=jd_4mo))
    assert np.all(mcmf0.separation(mcmf1) < ac.Angle("5arcsec"))


def test_sidereal_vs_solar_day():
    # Compute sidereal vs solar day length from time between transits of
    # the Sun and an arbitrary star.

    # 200 time steps over one month. Enough time to cover a full period
    ntimes = 200
    ts = Time("2025-01-01T00:00:00") + TimeDelta(np.linspace(0, 30, ntimes), format="jd")
    loc = lunarsky.MoonLocation.from_selenodetic(0, 56, 10.0)

    # Get Sun position and arbitrary star position
    sun_icrs = ac.get_sun(ts).transform_to("icrs")
    obj_icrs = ac.SkyCoord(ra="15d", dec="20d", frame="icrs")

    ltf = lunarsky.LunarTopo(location=loc, obstime=ts)
    sun_topo = sun_icrs.transform_to(ltf)
    obj_topo = obj_icrs.transform_to(ltf)

    # Measure period for each as double time between peak and trough
    sun_per = 2 * np.abs(
        ts[np.argmax(sun_topo.alt.deg)] - ts[np.argmin(sun_topo.alt.deg)]
    )
    obj_per = 2 * np.abs(
        ts[np.argmax(obj_topo.alt.deg)] - ts[np.argmin(obj_topo.alt.deg)]
    )

    assert np.isclose(sun_per.jd, 29.55, atol=0.1)
    assert np.isclose(obj_per.jd, 27.13, atol=0.1)


@pytest.mark.parametrize(
    "loc",
    [
        lunarsky.MoonLocation.from_selenodetic(lat, lon, 10.0)
        for lat, lon in [(27, 53), (10, 8)]
    ],
)
@pytest.mark.parametrize("toframe", ["lunartopo", "mcmf"])
def test_nearby_obj_transforms(toframe, loc):
    sun_gcrs = lunarsky.SkyCoord(ac.get_sun(jd_4mo))

    if toframe == "lunartopo":
        frame = lunarsky.LunarTopo(location=loc, obstime=jd_4mo)
    elif toframe == "mcmf":
        frame = lunarsky.MCMF(obstime=jd_4mo)

    res = sun_gcrs.transform_to(frame).transform_to(ac.GCRS(obstime=jd_4mo))
    assert_quantity_allclose(sun_gcrs.cartesian.xyz, res.cartesian.xyz)


@pytest.mark.parametrize(
    "Nt, Nl, success",
    [
        (2, 1, True),
        (1, 2, True),
        (2, 2, True),
        (2, 3, False),
        (3, 2, False),
    ],
)
def test_incompatible_shape_error(Nt, Nl, success):
    # Transforming to lunartopo

    latlons_sel = np.array(latlons_grid[:Nl]).T
    locs = lunarsky.MoonLocation.from_selenodetic(
        lat=latlons_sel[0] * un.deg, lon=latlons_sel[1] * un.deg
    )

    times = Time.now() + TimeDelta(np.linspace(0, 10, Nt), format="sec")

    if success:
        lunarsky.LunarTopo(location=locs, obstime=times)
    else:
        with pytest.raises(ValueError, match="non-scalar data and/or attributes"):
            lunarsky.LunarTopo(location=locs, obstime=times)


@pytest.mark.parametrize("fromframe", ["icrs", "mcmf"])
def test_incompatible_transform(fromframe):
    Nl = 3
    Nt = 1
    Ns = 5
    latlons_sel = np.array(latlons_grid[:Nl]).T
    locs = lunarsky.MoonLocation.from_selenodetic(
        lat=latlons_sel[0] * un.deg, lon=latlons_sel[1] * un.deg
    )

    times = Time.now() + TimeDelta(np.linspace(0, 10, Nt), format="sec")
    ltop = lunarsky.LunarTopo(location=locs, obstime=times)

    fromframe = ac.frame_transform_graph.lookup_name(fromframe)
    coo = fromframe().realize_frame(
        ac.SphericalRepresentation(
            lat=np.linspace(-10, 10, Ns) * un.deg,
            lon=np.linspace(30, 40, Ns) * un.deg,
            distance=1,
        )
    )
    src = lunarsky.SkyCoord(coo)
    with pytest.raises(IncompatibleShapeError):
        src.transform_to(ltop)


@pytest.mark.parametrize("ell", SELENOIDS)
def test_finite_vs_spherical(ell):
    # Transform MCMF coordinates with distance and without units
    # Check consistency with ellipsoid equatorial radius
    # Assumes infinite distance if no unit given, as astropy does.

    R0 = 404789  # km
    xyz = np.array([[R0, -R0], [0, 0], [0, 0]])
    with_units = lunarsky.SkyCoord(lunarsky.MCMF(*(xyz * un.km)))
    sans_units = lunarsky.SkyCoord(lunarsky.MCMF(*(xyz)))

    loc = lunarsky.MoonLocation.from_selenodetic(
        lon=180 * un.deg, lat=0, height=0, ellipsoid=ell
    )
    altaz_with_units = with_units.transform_to(lunarsky.LunarTopo(location=loc))
    with pytest.warns(exceptions.AstropyUserWarning, match="Coordinates do not "):
        altaz_sans_units = sans_units.transform_to(lunarsky.LunarTopo(location=loc))

    dists = R0 * un.km + np.array([1, -1]) * SELENOIDS[ell]._equatorial_radius
    assert np.all(altaz_with_units.distance == dists)
    assert_quantity_allclose(altaz_sans_units.distance, 1.0)


@pytest.mark.parametrize("ell", SELENOIDS)
@pytest.mark.parametrize("lat,lon", latlons_grid_nopole)
def test_topo_zenith_shift(ell, lat, lon):
    # Verify that a given source at zenith in one topo frame shifts
    # predictably when viewed from a topo frame with the same lat/lon but
    # different ellipsoid

    # Checking that the ellipsoid is interpreted correctly

    # This test is a little sketchy... will need to review this later.
    #   Some discrepancies for GRAIL23 selenoid when the source distance is large.
    #       Choosing 1000 km for now.
    #   Also fails near poles.

    # Comparing against the SPHERE ellipsoid. Test fails for this due to divide by zero
    if ell == "SPHERE":
        return

    lat *= un.deg
    lon *= un.deg

    loc0 = lunarsky.MoonLocation.from_selenodetic(
        lon, lat, ellipsoid="SPHERE"
    )  # For reference.
    loc1 = lunarsky.MoonLocation.from_selenodetic(
        lon, lat, ellipsoid=ell
    )  # Same lat/lon = different place for different ellipsoid

    # Zenith source at finite distance over loc0.
    src0 = lunarsky.SkyCoord(
        alt="90d",
        az="0d",
        distance=1e3 * un.km,
        frame=lunarsky.LunarTopo(location=loc0, obstime=Time.now()),
    )
    src1 = src0.transform_to(lunarsky.LunarTopo(location=loc1))

    # Law of sines:
    #   Geometry among zenith angle, mcmf station vector,
    #   and selenocentric vs. selenodetic latitudes

    R1 = loc1.mcmf.cartesian.norm()
    lat_cen = loc1.mcmf.spherical.lat
    lat_det = loc1.lat
    assert_quantity_allclose(
        src1.distance / np.sin((np.abs(lat_det - lat_cen)).rad),
        R1 / np.sin(src1.zen.rad),
        atol=1 * un.km,
    )



import numpy as np

from astropy.coordinates import ICRS, GCRS, EarthLocation, AltAz
from astropy.time import Time

from lunarsky import MoonLocation, SkyCoord, LunarTopo, MCMF

# Check that the changes to SkyCoord don't cause unexpected behavior.


def test_skycoord_transforms(grcat):
    # An EarthLocation object should still get copied over
    # under transformations.

    eloc = EarthLocation.from_geodetic(0.0, 10.0)

    altaz = grcat.transform_to(AltAz(location=eloc, obstime=Time.now()))

    assert altaz.location == eloc

    gcrs = altaz.transform_to(GCRS())

    assert gcrs.location == eloc

    icrs = altaz.transform_to(ICRS())

    assert icrs.location == eloc


def test_skycoord_with_lunar_frames():
    # Check that defining a SkyCoord with frames
    # lunartopo and mcmf works correctly.

    Nsrcs = 10
    alts = np.random.uniform(0, np.pi / 2, Nsrcs)
    azs = np.random.uniform(0, 2 * np.pi, Nsrcs)
    t0 = Time.now()
    loc = MoonLocation.from_selenodetic(0, 0)
    src = SkyCoord(
        alt=alts, az=azs, unit="rad", frame="lunartopo", obstime=t0, location=loc
    )

    assert src.location == loc
    assert isinstance(src.frame, LunarTopo)
    x, y, z = src.transform_to("mcmf").cartesian.xyz
    src2 = SkyCoord(x=x, y=y, z=z, frame="mcmf", obstime=t0, location=loc)

    assert isinstance(src2.frame, MCMF)

    icrs2 = src2.transform_to("icrs")
    icrs1 = src.transform_to("icrs")
    assert np.allclose(icrs2.ra.deg, icrs1.ra.deg, atol=1e-5)
    assert np.allclose(icrs2.dec.deg, icrs1.dec.deg, atol=1e-5)


def test_earth_and_moon():
    # Check that as I do transforms with both Earth and Moon positions,
    # the transform graph doesn't break.

    Nsrcs = 10
    alts = np.random.uniform(0, np.pi / 2, Nsrcs)
    azs = np.random.uniform(0, 2 * np.pi, Nsrcs)
    t0 = Time.now()
    loc = MoonLocation.from_selenodetic(0, 0)
    src = SkyCoord(
        alt=alts, az=azs, unit="rad", frame="lunartopo", obstime=t0, location=loc
    )

    eloc = EarthLocation.from_geodetic(0, 0)
    src2 = SkyCoord(
        alt=alts, az=azs, unit="rad", frame="altaz", obstime=t0, location=eloc
    )

    src.transform_to("icrs")
    src2.transform_to("icrs")



import numpy as np
import copy
from astropy import units as u
from astropy.units.quantity import QuantityInfoBase
from astropy.coordinates.angles import Longitude, Latitude
from astropy.coordinates.earth import GeodeticLocation
from astropy.coordinates.representation.geodetic import BaseGeodeticRepresentation
from astropy.coordinates.representation import (
    CartesianRepresentation,
)
from astropy.coordinates.attributes import Attribute

from .spice_utils import remove_topo

LUNAR_RADIUS = 1737.1e3  # m

__all__ = ["MoonLocation", "MoonLocationAttribute"]


class SPHERESelenodeticRepresentation(BaseGeodeticRepresentation):
    """Lunar ellipsoid as a sphere

    Radius defined by lunarsky.spice_utils.LUNAR_RADIUS
    """

    _equatorial_radius = LUNAR_RADIUS * u.m
    _flattening = 0.0


class GSFCSelenodeticRepresentation(BaseGeodeticRepresentation):
    """Lunar ellipsoid from NASA/GSFC "Planetary Fact Sheet"

    https://nssdc.gsfc.nasa.gov/planetary/factsheet/moonfact.html
    """

    _equatorial_radius = 1738.1e3 * u.m
    _flattening = 0.0012


class GRAIL23SelenodeticRepresentation(BaseGeodeticRepresentation):
    """Lunar ellipsoid defined by gravimetry of GRAIL data.

    https://doi.org/10.1007/s40328-023-00415-w
    """

    _equatorial_radius = 1737576.6 * u.m
    _flattening = 0.000305


class CE1LAM10SelenodeticRepresentation(BaseGeodeticRepresentation):
    """Lunar ellipsoid from Chang'e 1 laser altimetry.

    Rotation ellipsoid = CE-1-LAM-GEO

    https://doi.org/10.1007/s11430-010-4060-6
    """

    _equatorial_radius = 1737.632 * u.km
    _flattening = 1 / 973.463


# Define reference ellipsoids
SELENOIDS = {
    "SPHERE": SPHERESelenodeticRepresentation,
    "GSFC": GSFCSelenodeticRepresentation,
    "GRAIL23": GRAIL23SelenodeticRepresentation,
    "CE-1-LAM-GEO": CE1LAM10SelenodeticRepresentation,
}


class SelenodeticLocation(GeodeticLocation):
    """Rename GeodeticLocation class for clarity"""


def _check_ellipsoid(ellipsoid=None, default="SPHERE"):
    """Defaulting lunar ellipsoid"""
    if ellipsoid is None:
        ellipsoid = default
    if ellipsoid not in SELENOIDS:
        raise ValueError(f"Ellipsoid {ellipsoid} not among known ones ({SELENOIDS})")
    return ellipsoid


class MoonLocationInfo(QuantityInfoBase):
    """
    Container for meta information like name, description, format.  This is
    required when the object is used as a mixin column within a table, but can
    be used as a general way to store meta information.
    """

    _represent_as_dict_attrs = ("x", "y", "z", "ellipsoid")

    def _construct_from_dict(self, map):
        ellipsoid = map.pop("ellipsoid")
        out = self._parent_cls(**map)
        out.ellipsoid = ellipsoid
        return out

    def new_like(self, cols, length, metadata_conflicts="warn", name=None):
        """
        Return a new MoonLocation instance which is consistent with the
        input ``cols`` and has ``length`` rows.

        This is intended for creating an empty column object whose elements can
        be set in-place for table operations like join or vstack.

        Parameters
        ----------
        cols : list
            List of input columns
        length : int
            Length of the output column object
        metadata_conflicts : str ('warn'|'error'|'silent')
            How to handle metadata conflicts
        name : str
            Output column name

        Returns
        -------
        col : MoonLocation (or subclass)
            Empty instance of this class consistent with ``cols``
        """
        # Very similar to QuantityInfo.new_like, but the creation of the
        # map is different enough that this needs its own routine.
        # Get merged info attributes shape, dtype, format, description.
        attrs = self.merge_cols_attributes(
            cols, metadata_conflicts, name, ("meta", "format", "description")
        )
        # The above raises an error if the dtypes do not match, but returns
        # just the string representation, which is not useful, so remove.
        attrs.pop("dtype")
        # Make empty MoonLocation using the dtype and unit of the last column.
        # Use zeros so we do not get problems for possible conversion to
        # selenodetic coordinates.
        shape = (length,) + attrs.pop("shape")
        data = u.Quantity(
            np.zeros(shape=shape, dtype=cols[0].dtype), unit=cols[0].unit, copy=False
        )
        # Get arguments needed to reconstruct class
        map = {
            key: (data[key] if key in "xyz" else getattr(cols[-1], key))
            for key in self._represent_as_dict_attrs
        }
        out = self._construct_from_dict(map)
        # Set remaining info attributes
        for attr, value in attrs.items():
            setattr(out.info, attr, value)

        return out


class MoonLocation(u.Quantity):
    """
    Location on the Moon.

    There are two ``selenocentric'' coordinate systems in common use:
        - The Mean Axis / Polar axis (ME) system defines the z-axis as the mean rotational
    axis of the Moon, while the prime meridian is set by the mean Earth direction.
        - The Principal Axes (PA) frame is defined by the principal axes of the Moon.

    This class uses the ME frame.

    Positions may be defined in Cartesian (x, y, z) coordinates with respect to the
    center of mass of the Moon, or in ``selenodetic'' coordinates (longitude, latitude, height).

    Selenodetic coordinates are defined with respect to a reference ellipsoid. The default is a
    sphere of radius 1731.1e3 km, but other ellipsoids are available. See lunarsky.moon.SELENOIDS.

    See:
        "A Standardized Lunar Coordinate System for the Lunar Reconnaissance
        Orbiter and Lunar Datasets"
        LRO Project and LGCWG White Paper Version 5, 2008 October 1
        (https://lunar.gsfc.nasa.gov/library/LunCoordWhitePaper-10-08.pdf)

    Notes
    -----
    This class fits into the coordinates transformation framework in that it
    encodes a position on the `~astropy.coordinates.MCMF` frame.  To get a
    proper `~astropy.coordinates.MCMF` object from this object, use the ``mcmf``
    property.
    """

    _ellipsoid = "SPHERE"
    _location_dtype = np.dtype({"names": ["x", "y", "z"], "formats": [np.float64] * 3})
    _array_dtype = np.dtype((np.float64, (3,)))

    # Manage the set of defined ephemerides.
    # Class attributes only
    _inuse_stat_ids = []
    _avail_stat_ids = None
    _existing_locs = []
    _ref_count = []

    # This instance's station id(s)
    station_ids = []

    info = MoonLocationInfo()

    def __new__(cls, *args, **kwargs):
        # TODO: needs copy argument and better dealing with inputs.
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], MoonLocation):
            return args[0].copy()
        try:
            self = cls.from_selenocentric(*args, **kwargs)
        except (u.UnitsError, TypeError) as exc_selenocentric:
            try:
                self = cls.from_selenodetic(*args, **kwargs)
            except Exception as exc_selenodetic:
                raise TypeError(
                    "Coordinates could not be parsed as either "
                    "selenocentric or selenodetic, with respective "
                    'exceptions "{}" and "{}"'.format(exc_selenocentric, exc_selenodetic)
                )
        return self

    @classmethod
    def _set_site_id(cls, inst):
        """
        Set the station ID number and manage registry of defined stations.
        """
        if inst.isscalar:
            llh_arr = [
                (
                    inst.lon.deg.item(),
                    inst.lat.deg.item(),
                    inst.height.to_value("km").item(),
                )
            ]
            ncrds = 1
        else:
            llh_arr = zip(inst.lon.deg, inst.lat.deg, inst.height.to_value("km"))
            ncrds = inst.lon.size

        statids = []
        if cls._avail_stat_ids is None:
            cls._avail_stat_ids = list(range(999, 0, -1))

        if len(cls._avail_stat_ids) < ncrds:
            raise ValueError("Too many unique MoonLocation objects open at once.")

        for llh in llh_arr:
            lonlatheight = "_".join(
                ["{:.4f}".format(ll) for ll in llh] + [inst._ellipsoid]
            )
            if lonlatheight not in cls._existing_locs:
                new_stat_id = cls._avail_stat_ids.pop()
                cls._existing_locs.append(lonlatheight)
                statids.append(new_stat_id)
                cls._inuse_stat_ids.append(new_stat_id)
                cls._ref_count.append(1)
            else:
                ind = cls._existing_locs.index(lonlatheight)
                cls._ref_count[ind] += 1
                statids.append(cls._inuse_stat_ids[ind])
        inst.station_ids = statids
        return inst

    def _set_station_id(self):
        """
        Run classmethod for setting station IDs.

        Convenience function used for testing mostly
        """
        self.__class__._set_site_id(self)

    @classmethod
    def from_selenocentric(cls, x, y, z, unit=None):
        """
        Location on the Moon, initialized from selenocentric coordinates.

        Parameters
        ----------
        x, y, z : `~astropy.units.Quantity` or array_like
            Cartesian coordinates.  If not quantities, ``unit`` should be given.
        unit : `~astropy.units.UnitBase` object or None
            Physical unit of the coordinate values.  If ``x``, ``y``, and/or
            ``z`` are quantities, they will be converted to this unit.

        Raises
        ------
        astropy.units.UnitsError
            If the units on ``x``, ``y``, and ``z`` do not match or an invalid
            unit is given.
        ValueError
            If the shapes of ``x``, ``y``, and ``z`` do not match.
        TypeError
            If ``x`` is not a `~astropy.units.Quantity` and no unit is given.
        """
        if unit is None:
            try:
                unit = x.unit
            except AttributeError:
                raise TypeError(
                    "Selenocentric coordinates should be Quantities "
                    "unless an explicit unit is given."
                )
        else:
            unit = u.Unit(unit)

        if unit.physical_type != "length":
            raise u.UnitsError(
                "Selenocentric coordinates should be in " "units of length."
            )

        try:
            x = u.Quantity(x, unit, copy=False)
            y = u.Quantity(y, unit, copy=False)
            z = u.Quantity(z, unit, copy=False)
        except u.UnitsError:
            raise u.UnitsError(
                "Selenocentric coordinate units should all be " "consistent."
            )

        x, y, z = np.broadcast_arrays(x, y, z)
        struc = np.empty(x.shape, cls._location_dtype)
        struc["x"], struc["y"], struc["z"] = x, y, z
        inst = super().__new__(cls, struc, unit, copy=False)

        return inst

    @classmethod
    def from_selenodetic(cls, lon, lat, height=0.0, ellipsoid=None):
        """
        Location on the Moon, from latitude and longitude.

        Parameters
        ----------
        lon : `~astropy.coordinates.Longitude` or float
            Lunar East longitude.  Can be anything that initialises an
            `~astropy.coordinates.Angle` object (if float, in degrees).
        lat : `~astropy.coordinates.Latitude` or float
            Lunar latitude.  Can be anything that initialises an
            `~astropy.coordinates.Latitude` object (if float, in degrees).
        height : `~astropy.units.Quantity` or float, optional
            Height above reference sphere (if float, in meters; default: 0).
            The reference sphere is a sphere of radius 1737.1 kilometers,
            from the center of mass of the Moon.
        ellipsoid : str, optional
            Name of the reference ellipsoid to use (default: 'SPHERE').
            Available ellipsoids are:  'SPHERE', 'GRAIL23', 'CE-1-LAM-GEO'.
            See docstrings for classes in ELLIPSOIDS dictionary for references.

        Raises
        ------
        astropy.units.UnitsError
            If the units on ``lon`` and ``lat`` are inconsistent with angular
            ones, or that on ``height`` with a length.
        ValueError
            If ``lon``, ``lat``, and ``height`` do not have the same shape, or

        Notes
        -----

        latitude is defined relative to an equator 90 degrees
        off from the mean rotation axis. Longitude is defined
        relative to a prime meridian, which is itself given by
        the mean position of the "sub-Earth" point on the lunar surface.

        For the conversion to selenocentric coordinates, the ERFA routine
        ``gd2gce`` is used.  See https://github.com/liberfa/erfa

        """
        ellipsoid = _check_ellipsoid(ellipsoid, default=cls._ellipsoid)
        lon = Longitude(lon, u.degree, copy=False).wrap_at(180 * u.degree)
        lat = Latitude(lat, u.degree, copy=False)
        # don't convert to m by default, so we can use the height unit below.
        if not isinstance(height, u.Quantity):
            height = u.Quantity(height, u.m, copy=False)

        if not lon.shape == lat.shape:
            raise ValueError(
                "Inconsistent quantity shapes: {}, {}".format(
                    str(lon.shape), str(lat.shape)
                )
            )

        # get selenocentric coordinates. Have to give one-dimensional array.

        selenodetic = SELENOIDS[ellipsoid](lon, lat, height, copy=False)
        xyz = selenodetic.to_cartesian().get_xyz(xyz_axis=-1) << height.unit
        self = xyz.view(cls._location_dtype, cls).reshape(selenodetic.shape)
        self.ellipsoid = ellipsoid
        return self

    def __str__(self):
        return self.__repr__()

    def copy(self):
        # Necessary to preserve station_ids list
        c = super().copy()
        c.station_ids = self.station_ids
        return c

    def __copy__(self):
        # Ensure that the station_ids are copied as well under shallow copy
        obj = copy.copy(super())
        obj.station_ids = self.station_ids
        return obj

    def __del__(self):
        # Remove this MoonLocation's station_ids from _inuse_stat_ids and
        # locations from _existing_locs.
        # Also clear the corresponding frames from spice variable pool.
        for si, stat_id in enumerate(self.station_ids):
            try:
                ind = self.__class__._inuse_stat_ids.index(stat_id)
            except ValueError:
                continue
            count = self.__class__._ref_count[ind]
            if count <= 1:
                sid = self.__class__._inuse_stat_ids.pop(ind)
                self.__class__._existing_locs.pop(ind)
                self.__class__._avail_stat_ids.insert(0, sid)
                remove_topo(stat_id)
                self.__class__._ref_count.pop(ind)
            else:
                self.__class__._ref_count[ind] -= 1

    @property
    def selenodetic(self):
        """Convert to selenodetic coordinates."""
        return self.to_selenodetic()

    @property
    def ellipsoid(self):
        """The default ellipsoid used to convert to selenodetic coordinates."""
        return self._ellipsoid

    @ellipsoid.setter
    def ellipsoid(self, ellipsoid):
        self._ellipsoid = _check_ellipsoid(ellipsoid)

    def to_selenodetic(self, ellipsoid=None):
        """Convert to selenodetic coordinates (lat, lon, height).

        Height is in reference to the ellipsoid.

        Returns
        -------
        (lon, lat, height) : tuple
            The tuple contains instances of `~astropy.coordinates.Longitude`,
            `~astropy.coordinates.Latitude`, and `~astropy.units.Quantity`

        """
        ellipsoid = _check_ellipsoid(ellipsoid, default=self.ellipsoid)
        xyz = self.view(self._array_dtype, u.Quantity)
        llh = CartesianRepresentation(xyz, xyz_axis=-1, copy=False).represent_as(
            SELENOIDS[ellipsoid],
        )
        return SelenodeticLocation(
            Longitude(llh.lon, u.degree, wrap_angle=180.0 * u.degree, copy=False),
            Latitude(llh.lat, u.degree, copy=False),
            u.Quantity(llh.height, self.unit, copy=False),
        )

    @property
    def lon(self):
        """Longitude of the location"""
        return self.selenodetic[0]

    @property
    def lat(self):
        """Longitude of the location"""
        return self.selenodetic[1]

    @property
    def height(self):
        """Height of the location"""
        return self.selenodetic[2]

    # mostly for symmetry with selenodetic and to_selenodetic.
    @property
    def selenocentric(self):
        """Convert to a tuple with X, Y, and Z as quantities"""
        return self.to_selenocentric()

    def to_selenocentric(self):
        """Convert to a tuple with X, Y, and Z as quantities"""
        return (self.x, self.y, self.z)

    def get_mcmf(self, obstime=None):
        """
        Generates a `~lunarsky.mcmf.MCMF` object with the location of
        this object at the requested ``obstime``.

        Parameters
        ----------
        obstime : `~lunarsky.time.Time` or None
            The ``obstime`` to apply to the new `~lunarsky.mcmf.MCMF`, or
            if None, the default ``obstime`` will be used.

        Returns
        -------
        mcmf : `~lunarsky.mcmf.MCMF`
            The new object in the MCMF frame
        """
        # do this here to prevent a series of complicated circular imports
        from . import MCMF

        return MCMF(x=self.x, y=self.y, z=self.z, obstime=obstime)

    mcmf = property(
        get_mcmf,
        doc="""An `~astropy.coordinates.MCMF` object  with
                                     for the location of this object at the
                                     default ``obstime``.""",
    )

    def get_mcmf_posvel(self, obstime):
        """
        Calculate the MCMF position and velocity of this object at the
        requested ``obstime``.

        Parameters
        ----------
        obstime : `~astropy.time.Time`
            The ``obstime`` to calculate the GCRS position/velocity at.

        Returns
        --------
        obsgeoloc : `~astropy.coordinates.CartesianRepresentation`
            The GCRS position of the object
        obsgeovel : `~astropy.coordinates.CartesianRepresentation`
            The GCRS velocity of the object
        """
        # MCMF position
        mcmf_data = self.get_mcmf(obstime).data
        obspos = mcmf_data.without_differentials()
        obsvel = mcmf_data.differentials["s"].to_cartesian()
        return obspos, obsvel

    @property
    def x(self):
        """The X component of the selenocentric coordinates."""
        return self["x"]

    @property
    def y(self):
        """The Y component of the selenocentric coordinates."""
        return self["y"]

    @property
    def z(self):
        """The Z component of the selenocentric coordinates."""
        return self["z"]

    def __getitem__(self, item):
        result = super().__getitem__(item)
        if result.dtype is self.dtype:
            return result.view(self.__class__)
        else:
            return result.view(u.Quantity)

    def __array_finalize__(self, obj):
        super().__array_finalize__(obj)
        if hasattr(obj, "_ellipsoid"):
            self._ellipsoid = obj._ellipsoid

    def __len__(self):
        if self.shape == ():
            raise IndexError("0-d MoonLocation arrays cannot be indexed")
        else:
            return super().__len__()

    def _to_value(self, unit, equivalencies=[]):
        """Helper method for to and to_value."""
        # Conversion to another unit in both ``to`` and ``to_value`` goes
        # via this routine. To make the regular quantity routines work, we
        # temporarily turn the structured array into a regular one.
        array_view = self.view(self._array_dtype, np.ndarray)
        if equivalencies == []:
            equivalencies = self._equivalencies
        new_array = self.unit.to(unit, array_view, equivalencies=equivalencies)
        return new_array.view(self.dtype).reshape(self.shape)


class MoonLocationAttribute(Attribute):
    """
    A frame attribute that can act as a `~lunarsky.MoonLocation`.
    It can be created as anything that can be transformed to the
    `~lunarsky.MCMF` frame, but always presents as an `MoonLocation`
    when accessed after creation.

    Parameters
    ----------
    default : object
        Default value for the attribute if not provided
    secondary_attribute : str
        Name of a secondary instance attribute which supplies the value if
        ``default is None`` and no value was supplied during initialization.
    """

    def convert_input(self, value):
        """
        Checks that the input is a Quantity with the necessary units (or the
        special value ``0``).

        Parameters
        ----------
        value : object
            Input value to be converted.

        Returns
        -------
        out, converted : correctly-typed object, boolean
            Tuple consisting of the correctly-typed object and a boolean which
            indicates if conversion was actually performed.

        Raises
        ------
        ValueError
            If the input is not valid for this attribute.
        """

        if value is None:
            return None, False
        elif isinstance(value, MoonLocation):
            return value, False
        else:
            # we have to do the import here because of some tricky circular deps
            from . import MCMF

            if not hasattr(value, "transform_to"):
                raise ValueError(
                    '"{}" was passed into a '
                    "MoonLocationAttribute, but it does not have "
                    '"transform_to" method'.format(value)
                )
            mcmfobj = value.transform_to(MCMF())
            return mcmfobj.moon_location, True



import pytest
import copy
import numpy as np
import astropy.units as unit
from astropy.coordinates import (
    Longitude,
    Latitude,
)
from astropy.tests.helper import quantity_allclose
from lunarsky import MoonLocation, MoonLocationAttribute, MCMF
from lunarsky.moon import SELENOIDS


class TestsWithObject:
    # The following three functions have been adapted from corresponding
    # tests in astropy/coordinates/tests/test_earth.py
    def setup_method(self):
        # Check that the setup from different input formats works as expected.
        self.lon = Longitude(
            [0.0, 45.0, 90.0, 135.0, 180.0, -180, -90, -45],
            unit.deg,
            wrap_angle=180 * unit.deg,
        )
        self.lat = Latitude([+0.0, 30.0, 60.0, +90.0, -90.0, -60.0, -30.0, 0.0], unit.deg)
        self.h = unit.Quantity([0.1, 0.5, 1.0, -0.5, -1.0, +4.2, -11.0, -0.1], unit.m)
        self.location = MoonLocation.from_selenodetic(self.lon, self.lat, self.h)
        self.x, self.y, self.z = self.location.selenocentric

    def test_input(self):
        cartesian = MoonLocation(self.x, self.y, self.z)
        assert np.all(cartesian == self.location)
        cartesian = MoonLocation(self.x.value, self.y.value, self.z.value, self.x.unit)
        assert np.all(cartesian == self.location)
        spherical = MoonLocation(self.lon.deg, self.lat.deg, self.h.to(unit.m))
        # TODO this was just np.all(... == ...). Seeing floating point diff since v0.2.2
        assert quantity_allclose(
            spherical.mcmf.cartesian.xyz,
            self.location.mcmf.cartesian.xyz,
            atol=1e-9 * unit.m,
        )

    def test_invalid(self):
        # incomprehensible by either raises TypeError
        # Check error cases in setup.

        # TODO Include error messages in the check
        with pytest.raises(TypeError):
            MoonLocation(self.lon, self.y, self.z)

        # wrong units
        with pytest.raises(unit.UnitsError):
            MoonLocation.from_selenocentric(self.lon, self.lat, self.lat)
        # inconsistent units
        with pytest.raises(unit.UnitsError):
            MoonLocation.from_selenocentric(self.h, self.lon, self.lat)
        # floats without a unit
        with pytest.raises(TypeError):
            MoonLocation.from_selenocentric(self.x.value, self.y.value, self.z.value)
        # inconsistent shape
        with pytest.raises(ValueError):
            MoonLocation.from_selenocentric(self.x, self.y, self.z[:5])

        # inconsistent shape
        with pytest.raises(ValueError):
            MoonLocation.from_selenodetic(self.lon, self.lat[:5])

        # inconsistent units
        with pytest.raises(unit.UnitsError):
            MoonLocation.from_selenodetic(self.x, self.y, self.z)
        # inconsistent shape
        with pytest.raises(ValueError):
            MoonLocation.from_selenodetic(self.lon, self.lat, self.h[:5])

        # invalid ellipsoid
        with pytest.raises(ValueError):
            MoonLocation.from_selenodetic(self.lon, self.lat, self.h, ellipsoid="INVALID")

    def test_attributes(self):
        assert np.allclose(self.location.height, self.h)
        assert np.allclose(self.location.lon, self.lon)
        assert np.allclose(self.location.lat, self.lat)

    def test_mcmf_attr(self):
        mcmf = MCMF(x=self.x, y=self.y, z=self.z)
        assert np.allclose(mcmf.x, self.location.mcmf.x)
        assert np.allclose(mcmf.y, self.location.mcmf.y)
        assert np.allclose(mcmf.z, self.location.mcmf.z)


def test_moonlocation_attribute():
    # Make a MoonLocationAttribute in three ways
    # Confirm each looks like a typical MoonLocation

    testObj = TestsWithObject()
    testObj.setup_method()
    moonloc = testObj.location

    mlattr = MoonLocationAttribute()
    attr1, boo = mlattr.convert_input(None)
    assert attr1 is None
    assert not boo

    # If not None, this will look for a "transform_to"
    # method, which of course a string doesn't have.
    with pytest.raises(ValueError, match="was passed into a MoonLocationAttribute"):
        mlattr.convert_input("string")

    attr2, boo = mlattr.convert_input(moonloc.mcmf)
    assert np.all(attr2 == moonloc)


def test_moonlocation_copy():
    # Check that station_ids are copied properly
    loc0 = MoonLocation.from_selenodetic(lat=["-15d", "25d"], lon=["97d", "0d"])
    lcop = loc0.copy()
    assert lcop.station_ids == loc0.station_ids
    lcop2 = copy.copy(loc0)
    assert lcop2.station_ids == loc0.station_ids


def test_moonlocation_delete():
    # Check that making multiple instances of the same location raises the _ref_count
    # appropriately, and deleting them removes them correctly.
    before = copy.copy(MoonLocation._inuse_stat_ids)
    if MoonLocation._avail_stat_ids is not None:
        exp = before + MoonLocation._avail_stat_ids[-2:][::-1]
    else:
        exp = before + [1, 2]

    locs = []
    for ii in range(5):
        locs.append(MoonLocation.from_selenodetic(lat=["-15d", "25d"], lon=["97d", "0d"]))
        locs[-1]._set_station_id()

    check0 = copy.copy(MoonLocation._inuse_stat_ids)
    for ii in range(5, 0, -1):
        assert MoonLocation._ref_count[-1] == ii
        assert MoonLocation._inuse_stat_ids[-1] == check0[-1]
        locs.pop(ii - 1)

    check1 = copy.copy(MoonLocation._inuse_stat_ids)

    assert exp == check0
    assert check1 == before


@pytest.mark.parametrize("ell", SELENOIDS)
def test_station_ids(ell):
    # Check that when a MoonLocations are made, the appropriate station_ids are assigned.

    # Get whatever are already recorded in the class.
    orig_statids = copy.copy(MoonLocation._inuse_stat_ids)

    # Random positions with some repeats, in five groups
    lonlatheights = [
        [
            (294.67, -67.68, 15.23),
            (78.31, 51.60, 16.59),
            (335.27, 57.45, 27.32),
        ],
        [
            (335.27, 57.45, 27.32),
            (133.29, -80.63, 25.92),
            (323.28, -2.46, 4.74),
            (45.94, -25.17, 20.22),
            (326.40, 29.21, 9.67),
            (197.94, 3.92, 16.34),
        ],
        [(197.94, 3.92, 16.34)],
        [
            (197.94, 3.92, 16.34),
            (242.63, -20.86, 4.90),
            (70.30, -40.22, 29.91),
            (187.70, -85.44, 9.35),
            (129.55, 14.11, 0.24),
            (140.46, 4.19, 28.03),
            (232.22, 56.89, 15.22),
            (51.31, -30.92, 11.53),
            (55.59, -61.90, 3.98),
        ],
        [
            (307.76, 32.31, 29.88),
            (248.07, 2.89, 22.41),
            (349.94, -77.10, 14.27),
            (315.41, -72.71, 13.26),
            (173.39, 14.43, 3.37),
            (183.19, -83.01, 16.13),
            (258.53, -56.66, 29.75),
            (115.36, -44.42, 11.34),
            (270.61, 57.85, 26.16),
            (320.07, -75.06, 21.15),
            (216.74, 54.75, 1.92),
        ],
    ]

    all_pos = np.array(sum(lonlatheights, []))
    n_unique = np.unique(all_pos, axis=0).shape[0]

    locs = []
    locstrs = []

    for gp in lonlatheights:
        lons, lats, heights = np.array(gp).T
        locs.append(
            MoonLocation.from_selenodetic(
                lat=lats, lon=lons, height=heights, ellipsoid=ell
            )
        )
        locs[-1]._set_station_id()

    # Check that only unique positions got added
    added = len(MoonLocation._inuse_stat_ids) - len(orig_statids)
    assert added == n_unique

    statids = [loc.station_ids for loc in locs]

    locstrs = []
    for inst in locs:
        llhs = []
        if inst.isscalar:
            llh_arr = [
                (
                    inst.lon.deg.item(),
                    inst.lat.deg.item(),
                    inst.height.to_value("km").item(),
                )
            ]
        else:
            llh_arr = zip(inst.lon.deg, inst.lat.deg, inst.height.to_value("km"))
        for llh in llh_arr:
            llhs.append("_".join(["{:.4f}".format(ll) for ll in llh] + [ell]))
        locstrs.append(llhs)

    # Check that saved location strings correspond with station IDs in each instance
    for gi, gp in enumerate(statids):
        for sti, sid in enumerate(gp):
            ind = MoonLocation._inuse_stat_ids.index(sid)
            assert locstrs[gi][sti] == MoonLocation._existing_locs[ind]


def test_statid_management(cleanup_moonlocs):
    lats = np.random.uniform(-90, 90, 1000)
    lons = np.random.uniform(0, 360, 1000)

    for ii in range(2):
        loc = MoonLocation.from_selenodetic(
            lon=lons[ii * 500 : (ii + 1) * 500],
            lat=lats[ii * 500 : (ii + 1) * 500],
        )
        loc._set_station_id()
        assert len(loc.station_ids) == 500
        del loc

    bigloc = MoonLocation.from_selenodetic(lon=lons, lat=lats)
    with pytest.raises(ValueError, match="Too many unique MoonLocation"):
        bigloc._set_station_id()



from astropy.coordinates.baseframe import frame_transform_graph
from astropy.coordinates.transformations import FunctionTransformWithFiniteDifference
from astropy.coordinates import AltAz, ICRS, EarthLocation, Angle
from astropy.utils.data import download_files_in_parallel
import lunarsky
import lunarsky.spice_utils as spice_utils
from lunarsky.time import Time


import spiceypy as spice


def test_topo_frame_setup():
    # Check that the frame setup puts all the correct values in the kernel pool.

    latitude, longitude = 30, 25
    name, idnum, frame_dict, latlon = spice_utils.topo_frame_def(latitude, longitude)
    frame_strs = ["{}={}".format(k, v) for (k, v) in frame_dict.items()]
    spice.lmpool(frame_strs)

    for k, v in frame_dict.items():
        N, typecode = spice.dtpool(k)
        if typecode == "N":
            res = spice.gdpool(k, 0, 100)
            if len(res) == 1:
                res = [int(res[0])]
            if N > 1:
                v = [float(it) for it in v.split(" ")]
                res = res.tolist()
            else:
                res = res[0]
                v = int(v)
            assert v == res
        else:
            res = spice.gcpool(k, 0, 100)[0]
            v = v.replace("'", "")
            assert v == res


def test_kernel_paths():
    # Check that the correct kernel files are downloaded
    # Need to unhash the file names

    assert len(lunarsky.spice_utils.KERNEL_PATHS) == 5


def test_spice_earth(grcat):
    # Replace the ICRS->AltAz transform in astropy with one using SPICE.
    # Confirm that star positions are the same as with the original transform
    # to within the error due to relativistic aberration (~ 21 arcsec)

    # DSS-15, 399015
    lat, lon, height = 35.42185, -116.8871951, 973.211

    loc = EarthLocation.from_geodetic(lon, lat, height)
    t0 = Time.now()
    _J2000 = Time("J2000")

    altaz = grcat.transform_to(AltAz(location=loc, obstime=t0))

    # Make the Earth topo frame in spice.
    framename, idnum, frame_dict, latlon = lunarsky.spice_utils.topo_frame_def(
        lat, lon, moon=False
    )

    # One more kernel is needed for the ITRF93 frame.
    kname = "pck/earth_latest_high_prec.bpc"
    _naif_kernel_url = "https://naif.jpl.nasa.gov/pub/naif/generic_kernels"
    kurl = [_naif_kernel_url + "/" + kname]
    kernpath = download_files_in_parallel(
        kurl,
        cache=True,
        show_progress=False,
    )
    spice.furnsh(kernpath)

    frame_strs = ["{}={}".format(k, v) for (k, v) in frame_dict.items()]
    spice.lmpool(frame_strs)
    et = (t0 - _J2000).sec

    @frame_transform_graph.transform(FunctionTransformWithFiniteDifference, ICRS, AltAz)
    def icrs_to_ecef(icrs_coo, ecef_frame):
        mat = spice.pxform("J2000", framename, et)
        newrepr = icrs_coo.cartesian.transform(mat)

        return ecef_frame.realize_frame(newrepr)

    tr = frame_transform_graph.get_transform(ICRS, AltAz).transforms
    assert tr[0].func.__name__ == "icrs_to_ecef"  # Confirm new transform added correctly

    altaz2 = grcat.transform_to(AltAz(location=loc, obstime=t0))

    # Having done the transform, remove the spice transform from the graph
    frame_transform_graph.remove_transform(ICRS, AltAz, None)

    # Compare astropy topocentric coordinate transform results to
    assert all(altaz.separation(altaz2) < Angle("25arcsec"))
    # NOTE Large deviations in azimuth high altitudes.


def test_topo_kernel_setup():
    # Tests a single function in topo.py
    # Need to clear the kernel pool first

    spice.clpool()
    # Confirm no variables are loaded.
    assert not lunarsky.spice_utils.check_is_loaded("*")

    for filepath in lunarsky.spice_utils.KERNEL_PATHS:
        spice.furnsh(filepath)
    loc = lunarsky.MoonLocation.from_selenodetic(lat="30d", lon="20d")
    station_name, idnum, frame_specs, latlon = lunarsky.spice_utils.topo_frame_def(
        loc.lat.deg, loc.lon.deg, moon=True
    )
    statnum = idnum - 1301000
    lunarsky.topo._spice_setup(loc, [statnum])
    try:
        assert lunarsky.spice_utils.check_is_loaded("*{}*".format(idnum))
    finally:
        lunarsky.spice_utils.remove_topo(statnum)



import numpy as np
from astropy.coordinates import SkyCoord
import pytest
import warnings

from lunarsky.spice_utils import remove_topo
from lunarsky.moon import MoonLocation

# Ignore deprecation warnings from spiceypy
@pytest.fixture(autouse=True)
def ignore_representation_deprecation():
    warnings.filterwarnings(
        "ignore",
        message="Using or importing the ABCs from"
        "'collections' instead of from"
        "'collections.abc' is deprecated",
        category=DeprecationWarning,
    )


@pytest.fixture
def grcat():
    Nangs = 30
    ras = np.linspace(0, 360, Nangs)
    decs = np.linspace(-90 + 180 / Nangs, 90, Nangs, endpoint=False)
    ras, decs = map(np.ndarray.flatten, np.meshgrid(ras, decs))
    decs = np.insert(np.array([-90, 90]), 1, decs)
    ras = np.insert(np.array([0, 0]), 1, ras)

    stars = SkyCoord(ra=ras, dec=decs, unit="deg", frame="icrs")
    return stars


@pytest.fixture
def cleanup_moonlocs():
    # Reset the existing set of station IDs and corresponding kernel vars.
    yield

    for sid in MoonLocation._inuse_stat_ids:
        remove_topo(sid)

    MoonLocation._inuse_stat_ids = []
    MoonLocation._avail_stat_ids = None
    MoonLocation._existing_locs = []
    MoonLocation._ref_count = []



from .moon import *  # noqa
from .mcmf import *  # noqa
from .topo import *  # noqa
from .time import *  # noqa
from .sky_coordinate import SkyCoord  # noqa

from astropy.coordinates.baseframe import frame_transform_graph
from astropy.coordinates.attributes import EarthLocationAttribute

if "location" in frame_transform_graph.frame_attributes:
    frame_transform_graph.frame_attributes["location"] = EarthLocationAttribute(
        default=None
    )



import numpy as np
from astropy.coordinates import ICRS, EarthLocation
from astropy.time import TimeDelta
import pytest

from lunarsky import MoonLocation, SkyCoord, Time


@pytest.mark.parametrize("lat", np.linspace(-89, 89, 5))
@pytest.mark.parametrize("lon", np.linspace(0, 360, 5))
def test_sidereal_time_calculation(lat, lon):
    # Confirm that the ra of the zenith is close to the calculated LST.

    t0 = Time.now()
    loc = MoonLocation.from_selenodetic(lon, lat, 0)
    t0.location = loc

    Ntimes = 200
    Ndays = 28
    times = t0 + TimeDelta(np.linspace(0, Ndays, Ntimes) * 3600 * 24, format="sec")

    for tt in times:
        src = SkyCoord(alt="90d", az="0d", frame="lunartopo", obstime=tt, location=loc)
        lst = tt.sidereal_time("mean")
        assert np.isclose(lst.deg, src.transform_to(ICRS()).ra.deg, atol=1e-4)


def test_earthloc():
    lat = -30.0
    lon = 127.0

    t0 = Time(
        "2020-01-01T00:00:00", location=EarthLocation.from_geodetic(lon=lon, lat=lat)
    )
    t1 = Time("2020-01-01T00:00:00", location=(lon, lat))

    assert t0.location == t1.location



# Adjust the SkyCoord object to work with either MoonLocation or EarthLocation

from numpy import sqrt
from astropy.coordinates import (
    EarthLocationAttribute,
    EarthLocation,
    SkyCoord as aSkyCoord,
)
from astropy.coordinates import UnitSphericalRepresentation
from astropy.coordinates.baseframe import (
    BaseCoordinateFrame,
    frame_transform_graph,
    GenericFrame,
)
from astropy.coordinates.sky_coordinate_parsers import _get_frame_class
from astropy.constants import c as speed_of_light

from .moon import MoonLocation, MoonLocationAttribute
from .topo import LunarTopo


class SkyCoord(aSkyCoord):
    def __init__(self, *args, **kwargs):
        loc = kwargs.get("location", None)
        if isinstance(loc, MoonLocation):
            frame_transform_graph.frame_attributes["location"] = MoonLocationAttribute(
                default=None
            )
        elif isinstance(loc, EarthLocation):
            frame_transform_graph.frame_attributes["location"] = EarthLocationAttribute(
                default=None
            )
        super().__init__(*args, **kwargs)
        # Set the graph to its default
        frame_transform_graph.frame_attributes["location"] = EarthLocationAttribute(
            default=None
        )

    def transform_to(self, frame, merge_attributes=True):
        # a modified version of the corresponding astropy function.
        from astropy.coordinates.errors import ConvertError

        frame_kwargs = {}

        # Frame name (string) or frame class?  Coerce into an instance.
        try:
            frame = _get_frame_class(frame)()
        except Exception:
            pass

        if isinstance(frame, SkyCoord):
            frame = frame.frame  # Change to underlying coord frame instance

        if isinstance(frame, BaseCoordinateFrame):
            new_frame_cls = frame.__class__
            # Get frame attributes, allowing defaults to be overridden by
            # explicitly set attributes of the source if ``merge_attributes``.
            for attr in frame_transform_graph.frame_attributes:
                self_val = getattr(self, attr, None)
                frame_val = getattr(frame, attr, None)
                if frame_val is not None and not (
                    merge_attributes and frame.is_frame_attr_default(attr)
                ):
                    frame_kwargs[attr] = frame_val
                elif self_val is not None and not self.is_frame_attr_default(attr):
                    frame_kwargs[attr] = self_val
                elif frame_val is not None:
                    frame_kwargs[attr] = frame_val
        else:
            raise ValueError("Transform `frame` must be a frame name, class, or instance")

        # Hacky solution here -- Frames other than LunarTopo cannot accept a MoonLocation object
        # and can get confused by it. Do not pass along `location` unless certain it will work.
        moonloc_incompatible = not isinstance(frame, LunarTopo)
        graph_attrs = frame_transform_graph.frame_attributes
        if hasattr(self, "location"):
            loc = getattr(self, "location")
            if isinstance(loc, MoonLocation):
                if moonloc_incompatible:
                    frame_kwargs.pop("location")
            elif "location" in graph_attrs.keys() and isinstance(
                graph_attrs["location"], MoonLocationAttribute
            ):
                frame_transform_graph.frame_attributes[
                    "location"
                ] = EarthLocationAttribute(default=None)

        # Get the composite transform to the new frame
        trans = frame_transform_graph.get_transform(self.frame.__class__, new_frame_cls)
        if trans is None:
            raise ConvertError(
                "Cannot transform from {} to {}".format(
                    self.frame.__class__, new_frame_cls
                )
            )

        # Make a generic frame which will accept all the frame kwargs that
        # are provided and allow for transforming through intermediate frames
        # which may require one or more of those kwargs.
        generic_frame = GenericFrame(frame_kwargs)

        # Do the transformation, returning a coordinate frame of the desired
        # final type (not generic).
        new_coord = trans(self.frame, generic_frame)

        # Finally make the new SkyCoord object from the `new_coord` and
        # remaining frame_kwargs that are not frame_attributes in `new_coord`.
        for attr in set(new_coord.frame_attributes.keys()) & set(frame_kwargs.keys()):
            frame_kwargs.pop(attr)

        return self.__class__(new_coord, **frame_kwargs)

    def radial_velocity_correction(self, kind="barycentric", obstime=None, location=None):

        # Need to copy location parsing syntax first.
        # location validation
        timeloc = getattr(obstime, "location", None)
        if location is None:
            if self.location is not None:
                location = self.location
                if timeloc is not None:
                    raise ValueError(
                        "`location` cannot be in both the "
                        "passed-in `obstime` and this `SkyCoord` "
                        "because it is ambiguous which is meant "
                        "for the radial_velocity_correction."
                    )
            elif timeloc is not None:
                location = timeloc
            else:
                raise TypeError(
                    "Must provide a `location` to "
                    "radial_velocity_correction, either as a "
                    "SkyCoord frame attribute, as an attribute on "
                    "the passed in `obstime`, or in the method "
                    "call."
                )

        elif self.location is not None or timeloc is not None:
            raise ValueError(
                "Cannot compute radial velocity correction if "
                "`location` argument is passed in and there is "
                "also a  `location` attribute on this SkyCoord or "
                "the passed-in `obstime`."
            )

        # Use parent method if given an EarthLocation
        if isinstance(location, EarthLocation):
            return super().radial_velocity_correction(
                kind=kind, obstime=obstime, location=location
            )

        from astropy.coordinates.solar_system import get_body_barycentric_posvel
        from astropy.coordinates import solar_system_ephemeris

        solar_system_ephemeris.set("jpl")

        # With a MoonLocation:
        # Get ICRS (barycentric) cartesian position and velocity of the Earth
        pos_moon, v_moon = get_body_barycentric_posvel("moon", obstime)
        if kind == "barycentric":
            v_origin_to_moon = v_moon
        elif kind == "heliocentric":
            v_sun = get_body_barycentric_posvel("sun", obstime)[1]
            v_origin_to_moon = v_moon - v_sun
        else:
            raise ValueError(
                "`kind` argument to radial_velocity_correction must "
                "be 'barycentric' or 'heliocentric', but got "
                "'{}'".format(kind)
            )

        mcmf_p, mcmf_v = location.get_mcmf_posvel(obstime)
        # transforming to GCRS is not the correct thing to do here, since we don't want to
        # include aberration (or light deflection)? Instead, only apply parallax if necessary
        if self.data.__class__ is UnitSphericalRepresentation:
            targcart = self.mcmf.cartesian
        else:
            # skycoord has distances so apply parallax
            obs_mcmf_cart = pos_moon + mcmf_p
            mcmf_cart = self.mcmf.cartesian
            targcart = mcmf_cart - obs_mcmf_cart
            targcart /= targcart.norm()

        if kind == "barycentric":
            beta_obs = (v_origin_to_moon + mcmf_v) / speed_of_light
            gamma_obs = 1 / sqrt(1 - beta_obs.norm() ** 2)
            gr = location.gravitational_redshift(obstime)
            # barycentric redshift according to eq 28 in Wright & Eastmann (2014),
            # neglecting Shapiro delay and effects of the star's own motion
            zb = gamma_obs * (1 + targcart.dot(beta_obs)) / (1 + gr / speed_of_light) - 1
            return zb * speed_of_light
        else:
            # do a simpler correction ignoring time dilation and gravitational redshift
            # this is adequate since Heliocentric corrections shouldn't be used if
            # cm/s precision is required.
            return targcart.dot(v_origin_to_moon + mcmf_v)



import numpy as np
import os
import tempfile
from astropy.utils.data import download_files_in_parallel
from astropy.coordinates.matrix_utilities import rotation_matrix
from astropy.time import Time
import astropy.units as unit

import spiceypy as spice

from .mcmf import MCMF
from .data import DATA_PATH

_J2000 = Time("J2000")

TEMPORARY_KERNEL_DIR = tempfile.TemporaryDirectory()


def check_is_loaded(search):
    """
    Search the kernel pool variable names for a given string.
    """
    try:
        spice.gnpool(search, 0, 100)
    except (spice.support_types.SpiceyError):
        return False
    return True


def list_kernels():
    """
    List loaded kernels.

    Returns
    -------
    list of str
        Kernel names (file paths)
    list of str
        Corresponding kernel types
    """
    knames, ktypes = [], []
    for typ in ["spk", "fk", "tk", "pck", "lsk"]:
        for ii in range(spice.ktotal(typ)):
            dat = spice.kdata(ii, typ)
            knames.append(dat[0])
            ktypes.append(dat[1])
    return knames, ktypes


def furnish_kernels():
    kernel_names = [
        "pck/moon_pa_de421_1900-2050.bpc",
        "fk/satellites/moon_080317.tf",
        "fk/satellites/moon_assoc_me.tf",
    ]
    kernel_paths = [os.path.join(DATA_PATH, kn) for kn in kernel_names]
    for kp in kernel_paths:
        spice.furnsh(kp)

    # LSK and DE430 Kernels
    knames = ["lsk/naif0012.tls", "spk/planets/de430.bsp"]
    _naif_kernel_url = "https://naif.jpl.nasa.gov/pub/naif/generic_kernels/"
    kurls = [_naif_kernel_url + kname for kname in knames]
    paths = download_files_in_parallel(kurls, cache=True, show_progress=False)
    for kp in paths:
        kernel_paths.append(kp)
        spice.furnsh(kp)

    return kernel_paths


def lunar_surface_ephem(pos_x, pos_y, pos_z, station_num=98):
    """
    Make an SPK for the point on the lunar surface

    Creates a temporary file and furnishes from that.

    Parameters
    ----------
    pos_x, pos_y, pos_z: float
        MCMF frame cartesian position in km
    station_num: int
        Station number

    Returns
    -------
    int:
        Ephemeris ID number
    """
    point_id = 301000 + station_num

    ets = np.array([spice.str2et("1950-01-01"), spice.str2et("2150-01-01")])

    states = np.zeros((len(ets), 6))
    states[:, :3] = np.repeat([[pos_x], [pos_y], [pos_z]], len(ets), axis=1).T

    center = 301
    frame = "MOON_ME"
    degree = 1

    fname = os.path.join(TEMPORARY_KERNEL_DIR.name, f"lunar_points.bsp")
    if os.path.exists(fname):
        spice.unload(fname)
        handle = spice.spkopa(fname)
    else:
        handle = spice.spkopn(fname, "SPK_FILE", 0)
    spice.spkw09(
        handle,
        point_id,
        center,
        frame,
        ets[0],
        ets[-1],
        "0",
        degree,
        len(ets),
        states.tolist(),
        ets.tolist(),
    )
    spice.spkcls(handle)
    spice.furnsh(fname)

    return point_id


def topo_frame_def(latitude, longitude, station_num=98, moon=True):
    """
    Make a list of strings defining a topocentric frame. This can then be loaded
    with spiceypy.lmpool.
    """
    if moon:
        idnum = 1301000
        station_name = f"LUNAR-TOPO-{station_num}"
        relative = "MOON_ME"
    else:
        # Used in tests only
        idnum = 1399000
        station_name = f"EARTH-TOPO-{station_num}"
        relative = "ITRF93"

    # The DSS stations are built into SPICE, and they number up to 66.
    # We will call this station number 98.
    idnum += station_num
    fm_center_id = idnum - 1000000

    ecef_to_enu = np.matmul(
        rotation_matrix(-longitude, "z", unit="deg"),
        rotation_matrix(latitude, "y", unit="deg"),
    ).T
    # Reorder the axes so that X,Y,Z = E,N,U
    ecef_to_enu = ecef_to_enu[[2, 1, 0]]

    mat = " ".join(map("{:.7f}".format, ecef_to_enu.flatten()))

    fmt_strs = [
        "FRAME_{1}                     = {0}",
        "FRAME_{0}_NAME                = '{1}'",
        "FRAME_{0}_CLASS               = 4",
        "FRAME_{0}_CLASS_ID            = {0}",
        "FRAME_{0}_CENTER              = {2}",
        "OBJECT_{2}_FRAME              = '{1}'",
        "TKFRAME_{0}_RELATIVE          = '{3}'",
        "TKFRAME_{0}_SPEC              = 'MATRIX'",
        "TKFRAME_{0}_MATRIX            = {4}",
    ]

    frame_specs = [
        s.format(idnum, station_name, fm_center_id, relative, mat) for s in fmt_strs
    ]

    frame_dict = {}

    for spec in frame_specs:
        k, v = map(str.strip, spec.split("="))
        frame_dict[k] = v

    latlon = ["{:.4f}".format(l) for l in [latitude, longitude]]

    return station_name, idnum, frame_dict, latlon


def remove_topo(station_num):
    """Remove a lunar station, by number, from variable pool."""

    idnum = 1301000 + station_num
    fm_center_id = idnum - 1000000
    station_name = f"LUNAR-TOPO-{station_num}"

    fmt_vars = [
        "FRAME_{1}",
        "FRAME_{0}_NAME",
        "FRAME_{0}_CLASS",
        "FRAME_{0}_CLASS_ID",
        "FRAME_{0}_CENTER",
        "OBJECT_{2}_FRAME",
        "TKFRAME_{0}_RELATIVE",
        "TKFRAME_{0}_SPEC",
        "TKFRAME_{0}_MATRIX",
    ]

    frame_vars = [s.format(idnum, station_name, fm_center_id) for s in fmt_vars]

    # Handle a bug in spiceypy for older versions of numpy
    if np.str_ is None:
        return
    for var in frame_vars:
        spice.dvpool(var)

    # Ideally, one would also remove the ephemeris data for this station from
    # the lunar_points.bsp file. This doesn't seem to be possible. However,
    # the ephemeris _should_ be overwritten if the station_id is reused.


def earth_pos_mcmf(obstimes):
    """
    Get the position of the Earth in the MCMF frame.

    Using SPICE.

    Used for tests.
    """
    ets = (obstimes - Time("J2000")).sec
    earthpos = np.stack(
        [spice.spkpos("399", et, "MOON_ME", "None", "301")[0] for et in ets]
    )
    earthpos = unit.Quantity(earthpos.T, "km")
    return MCMF(*earthpos, obstime=obstimes)


KERNEL_PATHS = furnish_kernels()



"""Extension of astropy's Time class."""

import numpy as np
import astropy
from astropy.coordinates import EarthLocation, Longitude

from .moon import MoonLocation
import spiceypy as spice

__all__ = ["Time", "TimeDelta"]


class Time(astropy.time.Time):
    _appended_docstr = """
    Notes
    -----
    location can also be given as a `lunarsky.MoonLocation` object.
    """

    __doc__ = astropy.time.Time.__doc__ + _appended_docstr

    def __init__(
        self,
        val,
        val2=None,
        format=None,
        scale=None,
        precision=None,
        in_subfmt=None,
        out_subfmt=None,
        location=None,
        copy=False,
    ):

        super_loc = None
        if isinstance(location, (EarthLocation, tuple)):
            super_loc = location

        super().__init__(
            val,
            val2=val2,
            format=format,
            scale=scale,
            precision=precision,
            in_subfmt=in_subfmt,
            out_subfmt=out_subfmt,
            location=super_loc,
            copy=copy,
        )

        if isinstance(location, MoonLocation):
            self.location = location

    def sidereal_time(self, kind, longitude=None, model=None):
        # Currently returns the zenith RA as the LST.
        if self.location is None or isinstance(self.location, EarthLocation):
            return super().sidereal_time(kind, longitude=longitude, model=model)

        if model is not None:
            raise ValueError(
                "The 'model' keyword is not supported" "'for MoonLocation sidereal_times."
            )

        # From here on, proceed assuming longitude or location refer
        # to the selenodetic coordinate system. "self.location" must
        # be defined in order to get here.

        et = np.atleast_1d((self - Time("J2000")).sec)
        mats = np.array([spice.pxform("MOON_ME", "J2000", t) for t in et])

        # Zenith vector
        zvec = self.location.mcmf.cartesian.xyz
        uzvec = zvec / np.linalg.norm(zvec)

        newvec = np.dot(mats, uzvec)
        return Longitude(np.arctan2(newvec[..., 1], newvec[..., 0]).squeeze(), "rad")

    sidereal_time.__doc__ = (
        astropy.time.Time.sidereal_time.__doc__
        + """
        Notes
        -----
        If the location attribute is a `~lunarsky.MoonLocation` instance, the 'kind' keyword
        is ignored and the calculated sidereal time is given as the ICRS frame right ascension
        of zenith over the MoonLocation.
        """
    )

    def light_travel_time(self, *args, **kwargs):
        if isinstance(self.location, MoonLocation):
            raise ValueError(
                "Light travel time calculations are not" "yet supported for MoonLocation"
            )

        return super().light_travel_time(*args, **kwargs)


# For convenience in import statements, TimeDelta is also included in this namespace.
TimeDelta = astropy.time.TimeDelta



# -*- mode: python; coding: utf-8 -*
# Copyright (c) 2018 Radio Astronomy Software Group
# Licensed under the 3-clause BSD License

from setuptools import setup, find_packages
import io


with io.open("README.md", "r", encoding="utf-8") as readme_file:
    readme = readme_file.read()

setup_args = {
    "name": "lunarsky",
    "author": "Adam E. Lanman",
    "url": "https://github.com/aelanman/lunarsky",
    "download_url": "https://github.com/aelanman/lunarsky/archive/refs/tags/v0.1.1.tar.gz",
    "license": "BSD",
    "description": "Astropy support for selenocentric (Moon)"
    "reference frames and lunar surface observatories.",
    "long_description": readme,
    "long_description_content_type": "text/markdown",
    "package_dir": {"lunarsky": "lunarsky"},
    "packages": find_packages(),
    "use_scm_version": {
        "root": ".",
        "relative_to": __file__,
        "version_scheme": "post-release",
        "local_scheme": "no-local-version",
        "write_to": "lunarsky/version.py",
    },
    "include_package_data": True,
    "test_suite": "pytest",
    "tests_require": ["pytest"],
    "setup_requires": ["pytest-runner", "setuptools_scm"],
    "install_requires": ["numpy>=1.15", "astropy>=6.0.0", "spiceypy", "jplephem"],
    "classifiers": [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Astronomy",
    ],
    "keywords": "astronomy moon spice",
}

if __name__ == "__main__":
    setup(**setup_args)



import numpy as np
import warnings
from astropy.utils.decorators import format_doc
from astropy.coordinates.representation import (
    SphericalRepresentation,
    SphericalCosLatDifferential,
    UnitSphericalRepresentation,
    CartesianRepresentation,
)
from astropy.utils import check_broadcast, exceptions
from astropy.coordinates.baseframe import (
    BaseCoordinateFrame,
    base_doc,
    frame_transform_graph,
)
from astropy.coordinates.attributes import TimeAttribute
from astropy.time import Time
from astropy import units as un
from astropy.coordinates.baseframe import RepresentationMapping
from astropy.coordinates.transformations import FunctionTransformWithFiniteDifference
from astropy.coordinates.builtin_frames.icrs import ICRS

from .mcmf import MCMF
from .moon import MoonLocationAttribute
from .spice_utils import check_is_loaded, topo_frame_def, lunar_surface_ephem

import spiceypy as spice

_90DEG = 90 * un.deg
__all__ = ["LunarTopo"]

_J2000 = Time("J2000", scale="tt")

doc_components = """
    az : `~astropy.coordinates.Angle`, optional, must be keyword
        The Azimuth for this object (``alt`` must also be given and
        ``representation`` must be None).
    alt : `~astropy.coordinates.Angle`, optional, must be keyword
        The Altitude for this object (``az`` must also be given and
        ``representation`` must be None).
    distance : :class:`~astropy.units.Quantity`, optional, must be keyword
        The Distance for this object along the line-of-sight."""

doc_footer = """
    Other parameters
    ----------------
    obstime : `~astropy.time.Time`
        The time at which the observation is taken.  Used for determining the
        position and orientation of the Earth.
    location : `~lunarsky.MoonLocation`
        The location on the Moon.
    """


@format_doc(base_doc, components=doc_components, footer=doc_footer)
class LunarTopo(BaseCoordinateFrame):
    """
    An "East/North/Up" coordinate frame on the lunar surface, analogous to the
    AltAz frame in astropy.
    """

    frame_specific_representation_info = {
        SphericalRepresentation: [
            RepresentationMapping("lon", "az"),
            RepresentationMapping("lat", "alt"),
        ]
    }

    default_representation = SphericalRepresentation
    default_differential = SphericalCosLatDifferential

    obstime = TimeAttribute(default=_J2000)
    location = MoonLocationAttribute(default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def zen(self):
        """
        The zenith angle for this coordinate
        """
        return _90DEG.to(self.alt.unit) - self.alt


# -----------------
# Helper functions
# -----------------


def _spice_setup(locations, station_ids):
    for li, loc in enumerate(np.atleast_1d(locations)):
        sid = int(station_ids[li])
        frameloaded = check_is_loaded(f"FRAME_LUNAR-TOPO-{sid}")
        if not frameloaded:
            lunar_surface_ephem(
                loc.x.to_value("km"),
                loc.y.to_value("km"),
                loc.z.to_value("km"),
                station_num=sid,
            )  # Furnishes SPK for lunar surface point
            station_name, idnum, frame_specs, latlon = topo_frame_def(
                loc.lat.deg, loc.lon.deg, moon=True, station_num=sid
            )
            frame_strs = ["{}={}".format(k, v) for (k, v) in frame_specs.items()]
            spice.lmpool(frame_strs)


def make_transform(coo, toframe):
    ap_to_spice = {"icrs": ("J2000", 0), "mcmf": ("MOON_ME", 301)}
    # Get target frame and source frame names
    if isinstance(coo, LunarTopo):
        totopo = False
        frame_spice_name, frame_id = ap_to_spice[toframe.name]
        obstime = coo.obstime
        location = coo.location
    elif isinstance(toframe, LunarTopo):
        totopo = True
        frame_spice_name, frame_id = ap_to_spice[coo.name]
        obstime = toframe.obstime
        location = toframe.location

    if location is None:
        raise ValueError("location must be defined for LunarTopo transformations")

    # Initialize station_ids if not defined.
    if location.station_ids == []:
        location.__class__._set_site_id(location)

    # Make arrays
    ets = (obstime - _J2000).sec
    stat_ids = np.asarray(location.station_ids)
    shape_out = check_broadcast(coo.shape, ets.shape, location.shape)

    # Set up SPICE ephemerides and frame details
    _spice_setup(location, stat_ids)

    ets_ids = np.atleast_2d(np.stack(np.broadcast_arrays(ets, stat_ids)).T)

    coo_cart = coo.cartesian

    # Make rotation matrices
    mats = np.asarray(
        [
            spice.pxform(f"LUNAR-TOPO-{int(stat_id)}", frame_spice_name, et)
            for (et, stat_id) in ets_ids
        ]
    )
    if totopo:
        mats = np.linalg.inv(mats)

    # Check for unitspherical
    is_unitspherical = (
        isinstance(coo.data, UnitSphericalRepresentation)
        or coo.cartesian.x.unit == un.one
    )

    # If not unitspherical, shift by origin vector before rotating.
    if not is_unitspherical:
        # Make origin vector(s) in coo's frame.
        if totopo:
            origin_id = lambda n: int(frame_id)  # MCMF or ICRS frame origin
            target_id = lambda n: int(n) + 301000  # Station ID
            frame_name = lambda n: frame_spice_name

        else:
            origin_id = lambda n: int(n) + 301000
            target_id = lambda n: int(frame_id)
            frame_name = lambda n: f"LUNAR-TOPO-{int(n)}"

        orig_posvel = (
            np.asarray(
                [
                    spice.spkgeo(
                        target_id(stat_id), et, frame_name(stat_id), origin_id(stat_id)
                    )[0]
                    for (et, stat_id) in ets_ids
                ]
            )
            * un.km
        )
        coo_cart -= CartesianRepresentation((orig_posvel.T)[:3])

    newrepr = coo_cart.transform(mats).reshape(shape_out)

    if is_unitspherical:
        if not np.allclose(newrepr.norm(), 1.0):
            warnings.warn(
                "Coordinates do not all have unit magnitude, but will be treated as unit spherical,"
                "Define coordinates as Quantity or normalize to remove this warning.",
                exceptions.AstropyUserWarning,
            )
        newrepr = newrepr.represent_as(UnitSphericalRepresentation)

    return toframe.realize_frame(newrepr)


# -----------------
# Transformations
# -----------------
@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, ICRS, LunarTopo)
def icrs_to_lunartopo(icrs_coo, topo_frame):
    return make_transform(icrs_coo, topo_frame)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, LunarTopo, ICRS)
def lunartopo_to_icrs(topo_coo, icrs_frame):
    return make_transform(topo_coo, icrs_frame)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, MCMF, LunarTopo)
def mcmf_to_lunartopo(mcmf_coo, topo_frame):
    # TODO:
    #   > What if mcmf_coo and topo_frame have different obstimes?
    #   > What if location has obstime?
    return make_transform(mcmf_coo, topo_frame)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, LunarTopo, MCMF)
def lunartopo_to_mcmf(topo_coo, mcmf_frame):
    return make_transform(topo_coo, mcmf_frame)


@frame_transform_graph.transform(
    FunctionTransformWithFiniteDifference, LunarTopo, LunarTopo
)
def lunartopo_to_lunartopo(topo_coo, topo_frame):
    # TODO Should be possible to directly transform between different lunartopo frames.
    # For now, just go through MCMF
    return topo_coo.transform_to(MCMF(obstime=topo_coo.obstime)).transform_to(topo_frame)
